/****************************************************************************************
/*-- Copyright (c) 2017 by SAS Institute Inc., Cary, NC, USA. All Rights Reserved. --*/
/*

Model         	SAS Forecast Analyst Workbench 
Subject Area  	Three stage Modeling in SAS 9.4 using Custom Analysis   
Description		This file stores macros used in three stage modeling in SAS 9.4 
				
Created by 		DDPO Team
Created on 		July 2017

******************************************************************************************/


/******************************************************************************************
Name : 	 ddcf_ca_call_three_stage_mac
Purpose: This macro initiates signon process for parallel processing,
/*****************************************************************************************/


%macro ddcf_ca_call_three_stage_mac ;

%global thread_counter total_thread_no ddcf_total_tasks;


/* Counting the total number of tasks in the current forecast group. */
proc sql noprint;
	select count(*) into: ddcf_total_tasks
    from control.forecast_subgroups&plan_id.;
quit;

%put &ddcf_total_tasks.;

data _null_;
set config.plan_thread;
where plan_id=input("&plan_id",best12.);
call symputx('total_thread_no',THREAD_NUM);
stop;
run;

option autosignon=YES sascmd="!sascmd";


/* Call signon statements. Number of signons is equal to maximum number of threads
			   that user wants to execute in parallel. */
%ddcf_signon_statements;

/*	If total number of threads spawned are less than total tasks.  */
%if  &total_thread_no. <= &ddcf_total_tasks. %then %do;

	%do task_number = 1 %to &total_thread_no.;  

		/* This is the actual thread execution code in Rsubmit-EndRsubmit block. */
	   %ddcf_ca_3stg_subdata_thread_exec(thread_num= &task_number.);

	%end;
	%do task_number = %eval(&total_thread_no. + 1) %to &ddcf_total_tasks. ;  

		/* Iteration condition test: This involves determination of which of the
		threads have completed execution. This is the thread that's available
		for further processing. */
       %ddcf_iterate_flag_conditions(iteration_no= &task_number.);

		/* This is the actual thread execution code in Rsubmit-EndRsubmit block. */
       %ddcf_ca_3stg_subdata_thread_exec(thread_num= &thread_counter.);

 	%end; 

%end;
/* Total tasks are less than total number of threads spawned.*/
%else %do;

 	%do task_number = 1 %to &ddcf_total_tasks.;  

		/* This is the actual thread execution code in Rsubmit-EndRsubmit block. */
       %ddcf_ca_3stg_subdata_thread_exec(thread_num= &task_number.);

    %end;

%end;



%mend;

/******************************************************************************************
Name : 	 ddcf_ca_3stg_subdata_thread_exec
Purpose: To call main macro for three stage modeling in SAS 9.4. 
		 This code will be executed in parallel processing.
/*****************************************************************************************/

%macro ddcf_ca_3stg_subdata_thread_exec(thread_num = ) ;
/*Get the current library references into macro variable*/
proc sql noprint;
	create table existing_lib as
	select libname, path
	from dictionary.libnames
	where upcase(libname) in ("SDL", "CONTROL", "CONFIG", "SCRATCH", "ABT");
quit;

data _null_;
	set existing_lib end=eof;
	length libmac $10000;
	retain libmac;
	libmac=trim(libmac)||"libname "||trim(libname)||' "'||trim(path)||'";';
	call symput("libmac",libmac);
run;

/* The remote procesing code will start here. */

/* Counter of forecast group made as Remote server macro variable.*/

%syslput task_number = &task_number. / remote = t&thread_num._&plan_id.;

%syslput thread_num = &thread_num. / remote = t&thread_num._&plan_id.;

%syslput interval = &interval. / remote = t&thread_num._&plan_id.;

%syslput plan_id = &plan_id. / remote = t&thread_num._&plan_id.;

%syslput prj_def_no = &prj_def_no. / remote = t&thread_num._&plan_id.;

%syslput selection_method = &selection_method. / remote = t&thread_num._&plan_id.;

%syslput confidence_level = &confidence_level. / remote = t&thread_num._&plan_id.;

%syslput criterion = &criterion. / remote = t&thread_num._&plan_id.;

%syslput holdoutpct = &holdoutpct. / remote = t&thread_num._&plan_id.;

%syslput libmac= &libmac;
		  
rsubmit t&thread_num._&plan_id. 
   WAIT=NO
   cmacvar=t&thread_num._&plan_id.   ;
   	%ddcf_init;
	%unquote(&libmac);

   %ddcf_capture_log(logprefix = three_stage_&prj_def_no._&thread_num._&task_number.,log_id =&plan_id.); 
    
	%let syscc=0; 

	%ddcf_plan_spec;
	%ddcf_retrieve_hier_str_count;

	/*Purpose: This macro stores actual code for three stage modeling in SAS 9.4*/

	%macro ddcf_ca_three_stage_subdata_mac ;

		data _null_;
		current = input("&task_number.",best12.);
		set control.forecast_subgroups&plan_id. point = current;
		count = count(ktrim(kleft(complete_hierarchy))," ") + 1;
			call symputx('fcst_current_hier_count',count);
			call symputx('i',ktrim(kleft(substr(ts_datasets_nm,5))));
			call symputx('fcst_hierarchy_string',complete_hierarchy);
			call symputx('library_path',library_path);
			call symputx('data',ts_datasets_nm);
			call symputx('indep_data',ts_indep_nm);
			call symputx('control_dataset_nm',control_dataset_nm);
			call symputx('libref',libref);
			call symputx('current_hierarchy', current_hierarchy);
		stop;
		run;

		 /* Assigning libref to the hierarchy path obtained above. */
		libname temp "&library_path.";

		data _null_;
			comma_fcst_hierarchy_string = translate("&fcst_hierarchy_string.",',',' ');
	  		call symputx('comma_fcst_hierarchy_string',comma_fcst_hierarchy_string);
		  	comma_kpi_indep_string = translate("&kpi_indep_string.",',',' ');
		  	call symputx('comma_kpi_indep_string',comma_kpi_indep_string);
		run;

		%if &holdoutpct. > 0 %then %do;
		
			proc sql  noprint;
				create table work.get_holdout_sample_size  as select distinct &comma_fcst_hierarchy_string., count(*)as record_cnt,  
					floor(calculated record_cnt * &holdoutpct./100)  as holdout_sample 
				from temp.&data. group by  &comma_fcst_hierarchy_string.;
			quit;

			
			proc sql noprint;
				create table work.create_partition as select a.*, b.record_cnt, b.holdout_sample 
				from temp.&data. a, work.get_holdout_sample_size  b 
				where %do j = 1 %to &fcst_current_hier_count.;
								%let string&j. = %scan(&fcst_hierarchy_string.,&j.);
							%if &j.=1 %then %do;
								a.&&string&j.  = b.&&string&j.
							%end;
							%else %do;
								and	a.&&string&j.  = b.&&string&j.
							%end; 	 
					%end;
					;
			quit;

			
			data work.create_partition(drop=record_cnt cnt holdout_sample);
			set work.create_partition;
			by  &fcst_hierarchy_string.;
			retain cnt 0 ;
			if first.&current_hierarchy. then do;
			cnt=0;
			end;
			cnt=cnt+1;
			if cnt <= (record_cnt-holdout_sample) then partition_var=1;
			else partition_var=0;
			run;


			proc sort data=work.create_partition ;
			  by &fcst_hierarchy_string. start_dt;
			run;

		%end;
		%else %do;
			proc sort data=temp.&data. out=work.create_partition ;
			  by &fcst_hierarchy_string. start_dt;
			run;
		%end;

		proc sort data=temp.&indep_data. out=indep_data;
		  by &fcst_hierarchy_string. start_dt;
		run;

		/* missing values for independent variables are set based on the values specied in hpf_preferences table */
		%ddcf_ca_3stg_impute_indep_var;


		data temp.ca_indata_temp_&prj_def_no._&i;
		merge work.create_partition indep_data_tmp2;
		by &fcst_hierarchy_string. start_dt;
		run;


		/* Step 1: Timeseries decomposition */
		proc timeseries data=temp.ca_indata_temp_&prj_def_no._&i outdecomp=outcomp out=out1 maxerror=0;
		   decomp orig tcc sc sa sic / mode=add;
		   id start_dt interval=&interval. ;
		   by &fcst_hierarchy_string.;
		   var &kpi_variable_nm.;
		run;

		/* Generate de-seasonalized data */
		data temp.ca_outcomp_temp_&prj_def_no._&i.;
		set outcomp;
		if original ne . then do;
			adj_sc =sum(original, -sc);
		end;
		else adj_sc=.;
		run;


		/* Step 2: Analyze de-seasonal data to identify the impact of causal variables.  */
		proc sql;
			create table work.multi_reg_indata as select a.*, b.adj_sc, b.sc
			from temp.ca_indata_temp_&prj_def_no._&i a, temp.ca_outcomp_temp_&prj_def_no._&i. b
			where 	a.start_dt =b.start_dt
					%do j = 1 %to &fcst_current_hier_count.;
						%let string&j. = %scan(&fcst_hierarchy_string.,&j.);
							and	a.&&string&j.  = b.&&string&j.
					%end; 
					;
		quit;

		proc sort data=work.multi_reg_indata;
		by &fcst_hierarchy_string. start_dt;
		run;

		ods output ParameterEstimates=temp.ca_reg_param_est_temp_&prj_def_no._&i.;
		proc hpreg data=work.multi_reg_indata noclprint alpha=%sysevalf((100-&confidence_level.)/100);
		model adj_sc = &kpi_indep_string. ;
		%if %sysfunc(kupcase(%sysfunc(ktrim(%sysfunc(kleft("&selection_method.")))))) ne "NONE" %then %do; 
			selection method=&selection_method. ;
		%end;
		%if &holdoutpct. > 0 %then %do;
			partition role=partition_var(train='1' validate='0');
		%end;
		by &fcst_hierarchy_string.;

		%if %sysfunc(kupcase(%sysfunc(ktrim(%sysfunc(kleft("&selection_method.")))))) eq "LASSO" OR %sysfunc(kupcase(%sysfunc(ktrim(%sysfunc(kleft("&selection_method.")))))) eq "LAR" %then %do; 
			output out =temp.ca_reg_out_temp_&prj_def_no._&i. predicted=pred residual=residual  role=role copyvars=(_all_);
		%end;
		%else %do;
			output out =temp.ca_reg_out_temp_&prj_def_no._&i. predicted=pred residual=residual lcl=lcl ucl=ucl role=role copyvars=(_all_);
		%end;

		run;
		ods output close;



	/* Step 3: : Execute Proc HPF to forecast the residuals. */ 

		proc sort data=temp.ca_reg_out_temp_&prj_def_no._&i.;
		  	by &fcst_hierarchy_string. start_dt;
		run;

		proc hpf data=temp.ca_reg_out_temp_&prj_def_no._&i.  lead=&plan_no_of_fcst_period. outfor=temp.ca_res_forecast_temp_&prj_def_no._&i. maxerror=0; 
		   	id start_dt interval=&interval.; 
		   	forecast residual /criterion=&criterion.
			%if &holdoutpct. > 0 %then %do;
				holdoutpct=&holdoutpct.
			%end;
			; 
			by &fcst_hierarchy_string.;
		run; 

		/* Add causal impacts and seasonal components to residual forecasts and generate the final forecasted values. */
		proc sql;
		create table work.outfor as select a.*, 
			b.predict, sum(sc, pred, predict) as final_forecast
		from temp.ca_reg_out_temp_&prj_def_no._&i. a , temp.ca_res_forecast_temp_&prj_def_no._&i. b
		where a.start_dt =b.start_dt
				%do j = 1 %to &fcst_current_hier_count.;
					%let string&j. = %scan(&fcst_hierarchy_string.,&j.);
						and	a.&&string&j.  = b.&&string&j.
				%end; 
				;
		quit;


		data work.ca_outfor_temp;
		set work.outfor (keep=&fcst_hierarchy_string. start_dt &kpi_variable_nm. final_forecast );
		error=&kpi_variable_nm.-final_forecast;
		Lower =.;
		Upper=.;
		STD=.;
		format start_dt nldate.;
		rename &kpi_variable_nm.=actual
				final_forecast=predict;
		run;

		/* Generate upper, lower, std using proc hpfengine for external forecast */

		proc sort data=work.ca_outfor_temp;
		  	by &fcst_hierarchy_string. start_dt;
		run;

		proc hpfengine data = work.ca_outfor_temp lead=&plan_no_of_fcst_period. out = _NULL_
		   	outfor = temp.ca_outfor_temp_&prj_def_no._&i.
		   	outstat = temp.ca_outstat_temp_&prj_def_no._&i.
			globalselection = exmselect errorcontrol=(severity=(high)stage=(forecasting));
		   	id start_dt interval =&interval.;
		    forecast actual ;
		    external predict   ;
		    by  &fcst_hierarchy_string.;
		run; 

		data temp.ca_outfor_temp_&prj_def_no._&i.;
		set temp.ca_outfor_temp_&prj_def_no._&i.;
		where actual ne . or predict ne . ;
		run;

	%mend;
	/*	Purpose: Set missing values for independent variables */

	%macro ddcf_ca_3stg_impute_indep_var ;

	data work.HPF_Preferences;
	set config.hpf_preferences&plan_id.;
	where plan_id eq input("&plan_id.",best12.) and 
	 	&plan_root_rk. eq input("0",best12.) and 
	  	&plan_leaf_rk. eq input("0",best12.) and 
		kupcase(ktrim(kleft(option))) like "DIAG_INPUT_SETMISSING%"  ;
	run;

	data _null_ ;
	set work.HPF_Preferences;
	call execute('%let ' || ktrim(Option) || '=' || trim(Value));
	run;

	/* set missing values of indep variables */
	proc timedata data=work.indep_data out=work.indep_data_tmp1;
	    by &fcst_hierarchy_string.;
	    id start_dt interval=&interval. ;

	    %do indep_counter=1 %to &indep_num.;

			var &&indep_var&indep_counter /setmissing = &&diag_input_setmissing&indep_counter.;

		%end;
	run;
		
	/* In case value of independent variable is still missing for all periods for particular BY groups, set them to 0 to avoid error */

	%do cnt= 1 %to &fcst_current_hier_count.;

		%let hierarchy_var_rk&cnt = %sysfunc(kscan(&fcst_hierarchy_string., &cnt ));

		%put &&hierarchy_var_rk&cnt;


	%end;

	proc means data =work.indep_data_tmp1 noprint;
	by &fcst_hierarchy_string.;
	var &kpi_indep_string.;
	output out=summary ;
	run;

	data summary_n;
	set summary;
	where _stat_='N';
	run;

	proc sql;
		create table work.indep_data_tmp2 as select 
				%do cnt=1 %to &fcst_current_hier_count;
						a.&&hierarchy_var_rk&cnt,
				 %end;
		 		%do indep_counter=1 %to &indep_num.;
					(case when b.&&indep_var&indep_counter=0 then 0 else a.&&indep_var&indep_counter  end) as &&indep_var&indep_counter ,
				%end;
				start_dt
		from work.indep_data_tmp1 a, summary_n b
		where 	%do j = 1 %to &fcst_current_hier_count.;
					%let string&j. = %scan(&fcst_hierarchy_string.,&j.);
					%if &j=1 %then %do;
						a.&&string&j.  = b.&&string&j.
					%end;
					%else %do;
						and	a.&&string&j.  = b.&&string&j.
					%end; 
				%end;
			;
	quit;

	%mend;

	%ddcf_ca_three_stage_subdata_mac; 

	%ddcf_reset_log;
	%unquote(%nrstr(%sysrput) return&task_number = &syscc%str(;)); 
endrsubmit;

/* List of all t&thread_num._&plan_id. */
listtask _all_;


%mend;

/******************************************************************************************
Name : 	 ddcf_ca_three_stg_append_subdata
Purpose: To append the output of the subgroups created in parallel processing 
		 for three stage modeling in SAS 9.4 
/*****************************************************************************************/

%macro ddcf_ca_three_stg_append_subdata ;

%do hierarchy_counter =  0 %to %eval(&fcst_stat_hier_count_original. - 1) %by 1;

	%let hierarchy_counter = &hierarchy_counter.;

   	/* Setting current libref */
   	%let _hpf_current_libref = _hpf&hierarchy_counter.;

	data _null_;
   	set  &_hpf_current_libref..control_counter;
    	call symputx('current_hierarchy_data_count',hierarchy_count);
   	run;
    	    
	/* Appending outfor and outstat subdatasets to the library level single
	outfor and outstat. */
	%do data_counter = 0 %to &current_hierarchy_data_count. %by 1;

		%let data_counter = &data_counter. ;

		%if %sysfunc(exist(&_hpf_current_libref..ca_indata_temp_&prj_def_no._&data_counter.)) %then %do;
			%ddcf_append_data(data = &_hpf_current_libref..ca_indata_temp_&prj_def_no._&data_counter.,
			            base= &_hpf_current_libref..ca_indata_&prj_def_no.);
			%ddcf_set_rc(&syserr);
		%end;

		%if %sysfunc(exist(&_hpf_current_libref..ca_outcomp_temp_&prj_def_no._&data_counter.)) %then %do;
			%ddcf_append_data(data = &_hpf_current_libref..ca_outcomp_temp_&prj_def_no._&data_counter.,
			            base= &_hpf_current_libref..ca_outcomp_&prj_def_no.);
			%ddcf_set_rc(&syserr);
		%end;

		%if %sysfunc(exist(&_hpf_current_libref..ca_outfor_temp_&prj_def_no._&data_counter.))  %then %do;
			%ddcf_append_data(data = &_hpf_current_libref..ca_outfor_temp_&prj_def_no._&data_counter., 
		            base= &_hpf_current_libref..ca_outfor_&prj_def_no.);
			%ddcf_set_rc(&syserr);
		%end;

		%if %sysfunc(exist(&_hpf_current_libref..ca_reg_param_est_temp_&prj_def_no._&data_counter.))  %then %do;
			%ddcf_append_data(data = &_hpf_current_libref..ca_reg_param_est_temp_&prj_def_no._&data_counter., 
		            base= &_hpf_current_libref..ca_reg_param_est_&prj_def_no.);
			%ddcf_set_rc(&syserr);
		%end;

		%if %sysfunc(exist(&_hpf_current_libref..ca_reg_out_temp_&prj_def_no._&data_counter.))  %then %do;
			%ddcf_append_data(data = &_hpf_current_libref..ca_reg_out_temp_&prj_def_no._&data_counter., 
		            base= &_hpf_current_libref..ca_reg_out_&prj_def_no.);
			%ddcf_set_rc(&syserr);
		%end;

		%if %sysfunc(exist(&_hpf_current_libref..ca_res_forecast_temp_&prj_def_no._&data_counter.))  %then %do;
			%ddcf_append_data(data = &_hpf_current_libref..ca_res_forecast_temp_&prj_def_no._&data_counter., 
		            base= &_hpf_current_libref..ca_res_forecast_&prj_def_no.);
			%ddcf_set_rc(&syserr);
		%end;

		%if %sysfunc(exist(&_hpf_current_libref..ca_outstat_temp_&prj_def_no._&data_counter.))  %then %do;
			%ddcf_append_data(data = &_hpf_current_libref..ca_outstat_temp_&prj_def_no._&data_counter., 
		            base= &_hpf_current_libref..ca_outstat_&prj_def_no.);
			%ddcf_set_rc(&syserr);
		%end;

	%end;

	proc datasets library=&_hpf_current_libref. memtype=data nolist nodetails nowarn;
		delete 
			ca_indata_temp_&prj_def_no._:
			ca_outcomp_temp_&prj_def_no._: 
			ca_outfor_temp_&prj_def_no._:
			ca_reg_param_est_temp_&prj_def_no._:
			ca_reg_out_temp_&prj_def_no._ :
			ca_res_forecast_temp_&prj_def_no._ :
			ca_outstat_temp_&prj_def_no._:
		;
	
	run;
	quit;
	%ddcf_set_rc(&syserr);
%end;
%mend;
/******************************************************************************************
Name : 	 ddcf_create_ca_disagg_forecasts
Purpose: To generate disaggregated forecast for three stage modeling in SAS 9.4. 
/*****************************************************************************************/


%macro ddcf_create_ca_disagg_forecasts ;

data _null_;
set control.plan_hierarchy&plan_id. ;
call symputx('fcst_stat_hier_count_orig',fcst_stat_hier_count_orig);
run;

%let disaggregation_counter =  &fcst_stat_hier_count_orig.; 

%do hierarchy_counter =  &disaggregation_counter. %to %eval(&hierarchy_count_original. - 1) %by 1;

    %let hierarchy_counter = &hierarchy_counter.;

    /* Previous hierarchy number. */
    %let hierarchy_counter_previous = %eval(&hierarchy_counter. - 1);

    /* Setting current and previous counter librefs. */
    %let _hpf_current_libref = _hpf&hierarchy_counter.;
    %let _hpf_previous_libref = _hpf&hierarchy_counter_previous.;

    /* Current libref forecast hierarchy string. This is used as hierarchy class. */
   	data _null_;
   	set work.hierarchy;
   	if libref = "&_hpf_current_libref." then do;
     	call symputx('fcst_hierarchy_string',complete_hierarchy);
   	end;
   	else if libref = "&_hpf_previous_libref." then do;
     	call symputx('fcst_higher_hierarchy_string',complete_hierarchy);
   	end;
   	run;

    /* Get comma separated hierarchy strings for current and higher librefs. */

    data _null_;
    	fcst_hier_string = translate("&fcst_hierarchy_string.",',',' ');
		call symputx('fcst_hier_string',fcst_hier_string);
    run; 

	%put &fcst_hier_string. ;

	/* Number of hierarchy levels in the current complete string. */
    data _null_;
      	fcst_current_count =  count("&fcst_hier_string.",",") + 1;
      	call symputx('fcst_current_count',fcst_current_count);
    run;

	%put &fcst_current_count. ;

	 data _null_;
      	fcst_higher_hier_string = translate("&fcst_higher_hierarchy_string.",',',' ');
      	call symputx('fcst_higher_hier_string',fcst_higher_hier_string);
   	run; 

            %let predict_tab = &_hpf_previous_libref..ca_outfor_&prj_def_no.;

	/*Code for dimension disaggregation*/

	%if &SYSSCP. eq WIN %then %do;  /* Windows Operating System */
		%include "!sasroot\ddcfsrv\sasmisc\ddcf_dimension_disaggregation.sas";
	%end; 
	%else %do;  /* UNIX Operating System */
	   	%include "!sasroot/misc/ddcfsrv/ddcf_dimension_disaggregation.sas";
	%end; 


	%let allow_negative_ind=0;

	%ddcf_dimension_disaggregation( plan_id = &plan_id.,
	 			i_disagg_actuals_tab = &_hpf_current_libref..data  , 
		 		i_agg_actuals_tab = &_hpf_previous_libref..data, 
				 i_agg_predict_tab =&predict_tab. ,
				 o_disagg_predict_tab = &_hpf_current_libref..ca_outfor_&prj_def_no. ,
				 plan_kpi_aggregation_method =&kpi_Aggregation_metod. ,
				 kpi_var_nm =&kpi_variable_nm. ,
				 agg_lvl_count =%eval(&fcst_current_count. -1),
				 disagg_lvl_count =&fcst_current_count. ,
				 agg_hier_lvls_comma =%bquote(&fcst_higher_hier_string.) ,
				 disagg_hier_lvls_comma = %bquote(&fcst_hier_string.) ,
				 plan_leaf_lvls_and_dim_leaf_lvls = %bquote(&fcst_hier_string.), 
				 var_nm_of_actuals = &kpi_variable_nm. ,
				 var_nm_of_predicts = predict ,
				 allow_negative_ind = &allow_negative_ind.
			  	);


	proc sort data = &_hpf_current_libref..data out = work.sorted_data;
	  	by plan_id &fcst_hierarchy_string. ;
	run;


	data &_hpf_current_libref..ca_outfor_&prj_def_no.;
	set &_hpf_current_libref..ca_outfor_&prj_def_no.;
	plan_id=&plan_id.;
	run;


    proc sql noprint;
	 	create index outfor_index on 
           &_hpf_current_libref..ca_outfor_&prj_def_no.(plan_id,&fcst_hier_string.,start_dt);
    quit;

	/* Actuals are stored in outfor dataset from 'data' dataset. */
	data &_hpf_current_libref..ca_outfor_&prj_def_no.(rename = (&kpi_variable_nm. = actual));
	merge &_hpf_current_libref..ca_outfor_&prj_def_no.(in = a) work.sorted_data(in = b);
	   by plan_id &fcst_hierarchy_string. start_dt;
	   if a = 1 then output;
	run;

	data &_hpf_current_libref..ca_outfor_&prj_def_no.;
	set &_hpf_current_libref..ca_outfor_&prj_def_no.;
	   	error = actual - predict;
	  	Lower =.;
		Upper=.;
		STD=.;
	run; 
%end;

%mend;
